#define _CRT_SECURE_NO_DEPRECATE
#include <windows.h>
#include <stdio.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "vector.h"
#include "triangulate.h"
#include <time.h>
#include <io.h>
#include <fcntl.h> 

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

int debug_voronoi = 0;
int debug_shape = 0;
#define MAX_SHAPE 6

char shape_name[6][80] = {
	"square",
	"triangle",
	"house",
	"hexagon",
	"random",
	"mouse (press space)"
};

void set_point(int type, vec3 *point, unsigned int &num_point, int width, int height)
{
	switch (type)
	{
	case 0:
		// square
		num_point = 4;
		point[3].x = -width / 2;
		point[3].y = -height / 2;
		point[3].z = 0.0f;

		point[2].x = -width / 2;
		point[2].y = height / 2;
		point[2].z = 0.0f;

		point[1].x = width / 2;
		point[1].y = height / 2;
		point[1].z = 0.0f;

		point[0].x = width / 2;
		point[0].y = -height / 2;
		point[0].z = 0.0f;
		break;
	case 1:
		// triangle
		num_point = 3;
		point[0].x = -width / 2;
		point[0].y = -height / 2;
		point[0].z = 0;

		point[1].x = -width / 2;
		point[1].y = height / 2;
		point[1].z = 0;

		point[2].x = width / 2;
		point[2].y = height / 4;
		point[2].z = 0;
		break;

	case 2:
		// house
		num_point = 5;
		point[0].x = -width / 2;
		point[0].y = -height / 2;
		point[0].z = 0;

		point[1].x = width / 2;
		point[1].y = height / 2;
		point[1].z = 0;

		point[2].x = -width / 2;
		point[2].y = height / 2;
		point[2].z = 0;

		point[3].x = width / 2;
		point[3].y = -height / 2;
		point[3].z = 0.0f;

		point[4].x = 0;
		point[4].y = (height / 2) * -3;
		point[4].z = 0.0f;
		break;

	case 3:
		// hexagon
		num_point = 6;
		point[0].x = -width / 2;
		point[0].y = -height / 2;
		point[0].z = 0;

		point[1].x = width / 2;
		point[1].y = height / 2;
		point[1].z = 0;

		point[2].x = -width / 2;
		point[2].y = height / 2;
		point[2].z = 0;

		point[3].x = width / 2;
		point[3].y = -height / 2;
		point[3].z = 0.0f;

		point[4].x = 0;
		point[4].y = (height / 2) * -3;
		point[4].z = 0.0f;


		point[5].x = 0;
		point[5].y = (height / 2) * 3;
		point[5].z = 0.0f;
		break;
	case 4:
		num_point = 10;

		for (unsigned int i = 0; i < num_point; i++)
		{
			point[i].x = (rand() % width * 5) - width * 5 / 2;
			point[i].y = (rand() % height * 5) - height * 5 / 2;
			point[i].z = 0;
		}
		break;
	case 5:
		num_point = 0;
		break;
	}


}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
{
	static TCHAR szAppName[] = TEXT("Bowyer Watson");
	HWND         hwnd;
	MSG          msg;
	WNDCLASS     wndclass;

	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Program requires Windows NT!"),
			szAppName, MB_ICONERROR);
		return 0;
	}

	hwnd = CreateWindow(szAppName, TEXT("Bowyer Watson"),
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, hInstance, NULL);

	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}


void RedirectIOToConsole(int debug)
{
	if (debug)
	{
		int	hConHandle;
		long	lStdHandle;
		FILE	*fp;
		CONSOLE_SCREEN_BUFFER_INFO	coninfo;

		AllocConsole();
		HWND hwndConsole = GetConsoleWindow();

		ShowWindow(hwndConsole, SW_MAXIMIZE);
		// set the screen buffer to be big enough to let us scroll text
		GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), &coninfo);

		coninfo.dwSize.Y = 512;
		SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE), coninfo.dwSize);

		// redirect unbuffered STDOUT to the console
		lStdHandle = (intptr_t)GetStdHandle(STD_OUTPUT_HANDLE);
		hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);

		fp = _fdopen(hConHandle, "w");
		*stdout = *fp;
		setvbuf(stdout, NULL, _IONBF, 0);

		// redirect unbuffered STDIN to the console
		lStdHandle = (intptr_t)GetStdHandle(STD_INPUT_HANDLE);
		hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);

		fp = _fdopen(hConHandle, "r");
		*stdin = *fp;
		setvbuf(stdin, NULL, _IONBF, 0);

		// redirect unbuffered STDERR to the console
		lStdHandle = (intptr_t)GetStdHandle(STD_ERROR_HANDLE);
		hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);
		fp = _fdopen(hConHandle, "w");
		*stderr = *fp;
		setvbuf(stderr, NULL, _IONBF, 0);

		// make cout, wcout, cin, wcin, wcerr, cerr, wclog and clog point to console as well
		//ios::sync_with_stdio();

		//Fix issue on windows 10
		FILE *fp2 = freopen("CONOUT$", "w", stdout);
	}
	else
	{
		freopen("altEngine.log", "a", stdout);
		freopen("altEngine.log", "a", stderr);
	}
}


#define MAX_POINT 4096
vec3 point[MAX_POINT];
unsigned int num_point = 0;

vec3 triangle[MAX_TRIANGLE];
unsigned int num_triangle = 0;

extern COLORREF color_table[12];
extern HBRUSH brush_table[12];
extern HPEN pen_table[12];

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static int  cxClient, cyClient;
	HDC         hdc;
	PAINTSTRUCT ps;
	static float scale = 0.1f;
	static HBRUSH hBrush;
	static HPEN hPen;
	static POINT center;
	static POINT offset;
	static Triangulate triangulate;
	static POINT last_pos;

	switch (message)
	{
	case WM_CREATE:
	{
		srand(time(0));

		hBrush = CreateSolidBrush(RGB(255,0,0));
		hPen = CreatePen(PS_SOLID, 1, RGB(128, 0, 0));

		for (int i = 0; i < 12; i++)
		{
			brush_table[i] = CreateSolidBrush(color_table[i]);
			pen_table[i] = CreatePen(PS_SOLID, 5, color_table[i]);
		}

		//SetTimer(hwnd, WM_USER + 1, 1000, 0);
		RedirectIOToConsole(true);


		return 0;
	}
	case WM_SIZE:
	{
		cxClient = LOWORD(lParam);
		cyClient = HIWORD(lParam);
		static int once = 1;


		if (once)
		{
			offset.x = cxClient / 2;
			offset.y = cyClient / 2;
			once = 0;
		}

		set_point(debug_shape, point, num_point, cxClient, cyClient);
		center.x = cxClient / 2;
		center.y = cyClient / 2;
		return 0;
	}
	case WM_TIMER:
		InvalidateRect(hwnd, NULL, false);
		return 0;
	case WM_PAINT:
	{
		hdc = BeginPaint(hwnd, &ps);


		char output[128] = { 0 };
		char name[64] = { 0 };

		triangulate.debug_BowyerWatson(hdc, point, num_point, triangle, num_triangle, scale, offset);

		if (Triangulate::draw_mode <= MAX_MODE)
			sprintf(name, "%s", Triangulate::draw_names[Triangulate::draw_mode]);
		sprintf(output, "draw_mode %d debug point %d drawing %s shape %s scale %f offset %d %d",
			Triangulate::draw_mode,
			Triangulate::debug_point,
			name,
			shape_name[debug_shape],
			scale,
			offset.x,
			offset.y);
		TextOut(hdc, 50, 50, output, strlen(output));

		EndPaint(hwnd, &ps);
		return 0;
	}
	case WM_ERASEBKGND:
	{
		return DefWindowProc(hwnd, message, wParam, lParam);
	}
	case WM_MOUSEWHEEL:
	{
		short int zDelta = (short)HIWORD(wParam);

		if (zDelta > 0)
		{
			scale += 0.01f;
		}
		else
		{
			if (scale - 0.01f > 0)
				scale -= 0.01f;
		}
		InvalidateRect(hwnd, NULL, true);
		return 0;
	}
	case WM_KEYDOWN:
	{
		switch (wParam)
		{
		case VK_NEXT:
			Triangulate::draw_mode++;
			InvalidateRect(hwnd, NULL, true);
			if (Triangulate::draw_mode >= MAX_MODE)
				Triangulate::draw_mode = MAX_MODE;
			break;
		case VK_PRIOR:
			Triangulate::draw_mode--;
			if (Triangulate::draw_mode < 0)
				Triangulate::draw_mode = 0;
			InvalidateRect(hwnd, NULL, true);
			break;
		case VK_LEFT:
			Triangulate::debug_point--;
			if (Triangulate::debug_point < 0)
				Triangulate::debug_point = 0;
			InvalidateRect(hwnd, NULL, true);
			break;
		case VK_RIGHT:
			Triangulate::debug_point++;
			if (Triangulate::debug_point >= num_point)
				Triangulate::debug_point = num_point - 1;
			InvalidateRect(hwnd, NULL, true);
			break;
		case VK_UP:
			debug_shape++;
			if (debug_shape >= MAX_SHAPE)
				debug_shape = MAX_SHAPE - 1;
			set_point(debug_shape, point, num_point, cxClient, cyClient);
			InvalidateRect(hwnd, NULL, true);
			break;
		case VK_DOWN:
			debug_shape--;
			if (debug_shape < 0)
				debug_shape = 0;
			set_point(debug_shape, point, num_point, cxClient, cyClient);
			InvalidateRect(hwnd, NULL, true);
			break;
		case VK_HOME:
			debug_voronoi++;
			InvalidateRect(hwnd, NULL, true);
			break;
		case VK_END:
			debug_voronoi--;
			InvalidateRect(hwnd, NULL, true);
			break;
		case VK_SPACE:

			if (num_point < MAX_POINT)
			{
				point[num_point].x = (last_pos.x - offset.x) / scale;
				point[num_point].y = (last_pos.y - offset.y) / scale;
				if ( !(point[num_point - 1].x == point[num_point].x  && point[num_point - 1].y == point[num_point].y) )
				{
					printf("adding point %d (%f, %f)\r\n", num_point, point[num_point - 1].x, point[num_point - 1].y);
					num_point++;
				}
				else
				{
					printf("not adding duplicate point\r\n", num_point, point[num_point - 1].x, point[num_point - 1].y);
				}
			}
			Triangulate::debug_point = num_point - 1;

			InvalidateRect(hwnd, NULL, true);
			break;
		case VK_RETURN:
			num_point = 0;
			InvalidateRect(hwnd, NULL, true);
			break;
		}
		return 0;
	}
	case WM_MOUSEMOVE:
	{
		int	x, y;

		x = LOWORD(lParam);
		y = HIWORD(lParam);

		if (MK_LBUTTON & wParam)
		{
			offset.x += (x - last_pos.x);
			offset.y += (y - last_pos.y);
			InvalidateRect(hwnd, NULL, true);
		}

		last_pos.x = x;
		last_pos.y = y;
	}
	return 0;
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}


